theory Dependency_Graph
imports
  "Q-Restricted_Rewriting"
  Substitution
  Countable
begin

section {* Preliminaries *}

type_synonym ('f, 'v) terms = "('f, 'v) term set"

text {* Combined set of left-hand sides and right-hand sides of a TRS. *}
definition
  terms :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms"
where
  "terms R \<equiv> lhss R \<union> rhss R"

definition
  DG :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) trs \<Rightarrow> (('f, 'v)rule \<times> ('f,'v)rule) set"
where
  "DG P Q R \<equiv> {((s,t), (u,v)). (s, t) \<in> P
    \<and> (u, v) \<in> P \<and> (\<exists>\<sigma> \<tau>. (t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep Q R)^*
    \<and> s \<cdot> \<sigma> \<in> NF_terms Q \<and> u \<cdot> \<tau> \<in> NF_terms Q)}"

section {* A semantic version of cap *}

text {*Applying a substitution to every term of a given set.*}
abbreviation
  subst_apply_set :: "('f, 'v) terms \<Rightarrow> ('f, 'v,'w) gsubst \<Rightarrow> ('f, 'w) terms"
  (infixl "\<cdot>set" 60)
where
  "T \<cdot>set \<sigma> \<equiv> (\<lambda>t. t \<cdot> \<sigma>) ` T"

text {*The \emph{cap} of a term is the `upper' part which is definitely not
changed by rewriting. Parts which are potentially changed, are replaced by
fresh variables (represented by @{term "Var (Inl ())"} below).*}
fun
  cap :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms
    \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, unit + 'v) term"
where
  "cap R Q S (Var x) = (if (\<exists>\<sigma>. S \<cdot>set \<sigma> \<subseteq> NF_terms Q \<and> Var x \<cdot> \<sigma> \<notin> NF (qrstep Q R))
    then Var (Inl ())
    else Var (Inr x))"
| "cap R Q S (Fun f ts) = (
    if (\<exists>\<sigma> u. S \<cdot>set \<sigma> \<subseteq> NF_terms Q \<and> (Fun f ts \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O rqrstep Q R)
      then Var (Inl ())
      else Fun f (map (cap R Q S) ts))"

(*The substitution is needed in the proof of Lemma 3.8*)
declare size_simps[simp]
fun
  fresh_instances_subst :: "('f, unit + 'v) term \<Rightarrow> ('f, 'v,'w) gsubst \<Rightarrow> ('f, 'w) terms" ("[_]'__")
where
  "[Var (Inl ())]_\<sigma> = UNIV" |
  "[Var (Inr x) ]_\<sigma> = {Var x \<cdot> \<sigma>}" |
  "[Fun f ss    ]_\<sigma> = {Fun f ts | ts.
    length ts = length ss \<and> (\<forall>i<length ss. (ts!i) \<in> [ss!i]_\<sigma>)}"
declare size_simps[simp del]

abbreviation
  fresh_instances :: "('f, unit + 'v) term \<Rightarrow> ('f, 'v) terms" ("[_]")
where
  "[t] \<equiv> [t]_Var"

lemma fresh_instances_subst_not_empty[simp]:
  "[t]_\<sigma> \<noteq> {}"
proof (induct t)
  case (Var x) thus ?case by (cases x) auto
next
  case (Fun f ts)
  hence "\<forall>t\<in>set ts. \<exists>s. s \<in> [t]_\<sigma>" by auto
  from bchoice[OF this] obtain s
    where in_class: "\<forall>t\<in>set ts. s t \<in> [t]_\<sigma>" by auto
  let ?ss = "map (\<lambda>i. (s (ts!i))) [0..<length ts]"
  have len: "length ?ss = length ts" by simp
  from in_class have "\<forall>i<length ts. (s (ts!i)) \<in> [ts!i]_\<sigma>" by auto
  hence "\<forall>i<length ts. (?ss!i) \<in> [ts!i]_\<sigma>" by simp
  with len have "Fun f ?ss \<in> [Fun f ts]_\<sigma>" by auto
  thus ?case by best
qed

lemma fresh_instances_subst_subst: "[t]_\<tau> \<cdot>set \<sigma> \<subseteq> [t]_(\<tau> \<circ>s \<sigma>)"
proof (induct t arbitrary: \<sigma> \<tau>)
  case (Var x)
  show ?case
  proof (cases x)
    case (Inl n)
    have "[Var x]_\<tau> \<cdot>set \<sigma> = UNIV \<cdot>set \<sigma>" by (simp add: Inl)
    moreover have "[Var x]_(\<tau> \<circ>s \<sigma>) = UNIV" by (simp add: Inl)
    moreover have "UNIV \<cdot>set \<sigma> \<subseteq> UNIV" by auto
    ultimately show ?thesis by blast
  next
    case (Inr y) thus ?thesis by simp
  qed
next
  case (Fun f ts)
  show ?case
  proof
    fix t assume in_Fun: "t \<in> [Fun f ts]_\<tau> \<cdot>set \<sigma>"
    then obtain ss where len: "length ss = length ts" and "\<forall>i<length ts. (ss!i) \<in> [ts!i]_\<tau>"
      and t: "t = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by auto
    moreover from Fun have "\<forall>i<length ts. [ts!i]_\<tau> \<cdot>set \<sigma> \<subseteq> [ts!i]_(\<tau> \<circ>s \<sigma>)" by auto
    ultimately have "\<forall>i<length ts. (ss!i) \<cdot> \<sigma> \<in> [ts!i]_(\<tau> \<circ>s \<sigma>)" by auto
    with len show "t \<in> [Fun f ts]_(\<tau> \<circ>s \<sigma>)" unfolding t by auto
  qed
qed

lemma fresh_instances_subst: "[t] \<cdot>set \<sigma> \<subseteq> [t]_\<sigma>"
proof -
  have "[t]_Var \<cdot>set \<sigma> \<subseteq> [t]_(Var \<circ>s \<sigma>)" by (rule fresh_instances_subst_subst)
  thus ?thesis by simp
qed

definition
  is_ecap ::
    "(('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, unit + 'v) term)
      \<Rightarrow> bool"
where
  "is_ecap ecap \<equiv> \<forall>R Q S t f ts.
    [cap R Q S t] \<subseteq> [ecap R Q S t]
    \<and> ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}"

lemma is_ecapI[intro]:
  assumes "\<And>R Q S t f ts. [cap R Q S t] \<subseteq> [ecap R Q S t]
    \<and> ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}"
  shows "is_ecap ecap"
using assms unfolding is_ecap_def by auto

lemma is_ecapE[elim]:
  assumes "is_ecap ecap"
    and "\<lbrakk>[cap R Q S t] \<subseteq> [ecap R Q S t];
      ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}\<rbrakk> \<Longrightarrow> P"
  shows P
using assms unfolding is_ecap_def by simp

lemma is_ecap_cap[intro]: "is_ecap cap"
proof
  fix R Q S t f ts
    show "[cap R Q S t] \<subseteq> [cap R Q S t]
      \<and> cap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (cap R Q S) ts)}"
  by (induct t) simp_all
qed

lemma UNIV_term_not_singleton[intro]: "(UNIV::('f,'v) terms) = {t} \<Longrightarrow> False"
  by (induct t) auto

lemma UNIV_fresh_instances_substE:
  assumes "UNIV = [t]_\<sigma>" and "t = Var (Inl ()) \<Longrightarrow> P" shows P
proof (cases t)
  case (Var x) show ?thesis using assms unfolding Var by (cases x) auto
next
  case (Fun f ts) show ?thesis using assms unfolding Fun by auto
qed

lemma Var_Inr_fresh_instancesE:
  assumes "[Var (Inr x)] \<subseteq> [t]" and "t = Var (Inl ()) \<Longrightarrow> P"
    and "t = Var (Inr x) \<Longrightarrow> P"
  shows P
proof (cases t)
  case (Var y) show ?thesis using assms unfolding Var by (cases y) auto
next
  case (Fun f ts) show ?thesis using assms unfolding Fun by auto
qed

lemma Fun_fresh_instancesE:
  assumes "Fun f ts \<in> [t]"
    and "t = Var (Inl ()) \<Longrightarrow> P"
    and "\<And>ss. \<lbrakk>t = Fun f ss;length ts = length ss; \<forall>i<length ss. ts!i \<in> [ss!i]\<rbrakk> \<Longrightarrow> P"
  shows P
proof (cases t)
  case (Var x) show ?thesis using assms unfolding Var by (cases x) auto
next
  case (Fun g ss) show ?thesis using assms unfolding Fun by auto
qed

lemma fresh_instances_Var_fresh: "t \<in> [Var (Inl ())]_\<sigma>" by simp

lemma fresh_instances_Var_old:
  assumes "t \<in> [Var (Inr x)]_\<sigma>" and "t = Var x \<cdot> \<sigma> \<Longrightarrow> P" shows P
using assms by simp

lemma equiv_class_Fun:
  assumes "t \<in> [Fun f ts]_\<sigma>"
    and "\<And>ss. \<lbrakk>t = Fun f ss; length ss = length ts; \<forall>i<length ts. ss!i \<in> [ts!i]_\<sigma>\<rbrakk> \<Longrightarrow> P"
  shows P
using assms by auto


lemma cap_Var_fresh:
  assumes "is_ecap ecap" and "cap R Q S (Var x) = Var (Inl ())"
  shows "ecap R Q S (Var x) = Var (Inl ())"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  show "ecap R Q S (Var x) = Var (Inl ())"
  proof (cases "cap R Q S (Var x) = Var (Inl ())")
    case True
    from cap[unfolded True] show ?thesis using UNIV_fresh_instances_substE[of _ Var] by auto
  next
    case False
    hence "cap R Q S (Var x) = Var (Inr x)" by (auto simp: split_ifs)
    with assms have False by auto
    thus ?thesis by simp
  qed
qed

lemma is_ecap_Var:
  assumes "is_ecap ecap" and "ecap R Q S (Var x) = Var (Inr x)"
  shows "cap R Q S (Var x) = Var (Inr x)"
proof -
  from assms have "ecap R Q S (Var x) \<noteq> Var (Inl ())" by auto
  with cap_Var_fresh[OF assms(1)]
    have "cap R Q S (Var x) \<noteq> Var (Inl ())" by blast
  thus ?thesis by (simp add: split_ifs)
qed

lemma cap_Var_old:
  assumes "is_ecap ecap" and "cap R Q S (Var x) = Var (Inr x)"
  shows "ecap R Q S (Var x) = Var (Inl ()) \<or> ecap R Q S (Var x) = Var (Inr x)"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  from Var_Inr_fresh_instancesE[OF this[unfolded assms]] show ?thesis by auto
qed

lemma cap_Var_cases:
  assumes "cap R Q S (Var x) = Var (Inr x) \<Longrightarrow> P"
    and "cap R Q S (Var x) = Var (Inl ()) \<Longrightarrow> P"
  shows P
using assms by (auto simp: split_ifs)

lemma ecap_Var_not_fresh:
  assumes "is_ecap ecap" and "ecap R Q S (Var x) \<noteq> Var (Inl ())"
  shows "ecap R Q S (Var x) = Var (Inr x)"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  show ?thesis
  proof (rule cap_Var_cases)
    assume "cap R Q S (Var x) = Var (Inr x)"
    from cap_Var_old[OF assms(1) this] and assms show ?thesis by simp
  next
    assume "cap R Q S (Var x) = Var (Inl ())"
    from cap_Var_fresh[OF assms(1) this] and assms show ?thesis by simp
  qed
qed

lemma cap_Fun_cases:
  assumes "cap R Q S (Fun f ts) = Var (Inl ()) \<Longrightarrow> P"
    and "cap R Q S (Fun f ts) = Fun f (map (cap R Q S) ts) \<Longrightarrow> P"
  shows P
using assms unfolding cap.simps split_ifs by blast

lemma cap_Fun_fresh:
  assumes "is_ecap ecap" and "cap R Q S (Fun f ts) = Var (Inl ())"
  shows "ecap R Q S (Fun f ts) = Var (Inl ())"
using assms(1) proof
  assume "[cap R Q S (Fun f ts)] \<subseteq> [ecap R Q S (Fun f ts)]"
  from UNIV_fresh_instances_substE[of _ Var] and this[unfolded assms] show ?thesis by auto
qed

lemma ecap_Fun:
  assumes "is_ecap ecap"
  shows "ecap R Q S (Fun f ts) = Var (Inl ())
    \<or> ecap R Q S (Fun f ts) = Fun f (map (ecap R Q S) ts)"
using assms by auto

(*FIXME: move *)
lemma subst_restrict_Int[simp]: "(\<sigma> |s V )|s W = \<sigma> |s (V \<inter> W)"
  by (rule ext) (simp add: subst_restrict_def)

lemma 
  assumes "x \<noteq> y"
  shows "Var x \<notin> UNIV \<cdot>set (\<lambda>_. Var y)" (is "_ \<notin> UNIV \<cdot>set ?\<sigma>")
proof
  assume "Var x \<in> UNIV \<cdot>set ?\<sigma>"
  then obtain t where "Var x = t \<cdot> ?\<sigma>" by auto
  with assms show False by (induct t) auto
qed

(* Lemma 3.8 of René's Thesis *)
lemma ecap_steps:
  assumes ecap: "is_ecap ecap"
    and step: "(t \<cdot> \<sigma>, u) \<in> (qrstep Q R)^*"
    and NF: "S \<cdot>set \<sigma> \<subseteq> NF_terms Q"
  shows "u \<in> [ecap R Q S t]_\<sigma>"
using step proof (induct t arbitrary: u)
  case (Var x)
  show ?case
  proof (cases "Var x \<cdot> \<sigma> \<in> NF (qrstep Q R)")
    case True
    show ?thesis
    proof (cases rule: cap_Var_cases)
      assume 1: "cap R Q S (Var x) = Var (Inl ())"
      (*this case is the reason for the subst in fresh_instances_subst.
      Otherwise we would have to prove "u : UNIV \<cdot>set \<sigma>", which does not
      hold.*)
      from cap_Var_fresh[OF ecap this] show ?thesis by auto
    next
      assume "cap R Q S (Var x) = Var (Inr x)"
      from cap_Var_old[OF ecap this]
        have "ecap R Q S (Var x) = Var (Inl ()) \<or> ecap R Q S (Var x) = Var (Inr x)" .
      thus ?thesis
      proof
        assume "ecap R Q S (Var x) = Var (Inl ())"
        thus ?thesis by auto
      next
        assume "ecap R Q S (Var x) = Var (Inr x)"
        moreover from Var and True have "u = Var x \<cdot> \<sigma>" by (induct rule: rtrancl.induct) auto
        ultimately show ?thesis by auto
      qed
    qed
  next
    case False
    with NF have "cap R Q S (Var x) = Var (Inl ())" by auto
    from cap_Var_fresh[OF ecap this] show ?thesis by auto
  qed
next
  case (Fun f ss)
  have "(\<exists>us. length us = length ss
    \<and> u = Fun f us \<and> (\<forall>i<length ss. ((ss ! i) \<cdot> \<sigma>, us ! i) \<in> (qrstep Q R)^*))
    \<or> (Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O (rqrstep Q R) O (qrstep Q R)^*"
    using qrsteps_rqrstep_cases[of f "map (\<lambda>t. t \<cdot> \<sigma>) ss" u Q R,
      OF Fun(2)[simplified], unfolded length_map]
      using nth_map[of _ ss "\<lambda>t. t \<cdot> \<sigma>"] by auto
  thus ?case
  proof
    assume "(Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O (rqrstep Q R) O (qrstep Q R)^*"
    then obtain v where "(Fun f ss \<cdot> \<sigma>, v) \<in> (qrstep Q R)^* O (rqrstep Q R)" by auto
    with NF
    have "cap R Q S (Fun f ss) = Var (Inl ())" by auto
    from cap_Fun_fresh[OF ecap this] show ?thesis by auto
  next
    assume "\<exists>us. length us = length ss
      \<and> u = Fun f us \<and> (\<forall>i<length ss. ((ss ! i) \<cdot> \<sigma>, us ! i) \<in> (qrstep Q R)^*)"
    then obtain us where len: "length us = length ss" and u: "u = Fun f us"
      and steps: "\<forall>i<length ss. ((ss!i)\<cdot>\<sigma>, us!i) \<in> (qrstep Q R)^*" by auto
    with Fun have args: "\<forall>i<length ss. us!i \<in> [ecap R Q S (ss!i)]_\<sigma>" by auto
    from ecap_Fun[OF ecap]
      have "ecap R Q S (Fun f ss) = Var (Inl ())
        \<or> ecap R Q S (Fun f ss) = Fun f (map (ecap R Q S) ss)" .
    thus ?thesis
    proof
      assume "ecap R Q S (Fun f ss) = Var (Inl ())" thus ?thesis by auto
    next
      assume ecap_map: "ecap R Q S (Fun f ss) = Fun f (map (ecap R Q S) ss)"
      from len and args have "Fun f us \<in> [ecap R Q S (Fun f ss)]_\<sigma>"
        unfolding ecap_map by auto
      thus ?thesis unfolding u .
    qed
  qed
qed

fun class_to_term_intern :: "int \<Rightarrow> ('f,unit + 'v)term \<Rightarrow> int \<times> (('f,int + 'v)term)"
  where "class_to_term_intern i (Fun f ts) = (
       let (k,ss) = foldr
        (\<lambda> t (j,ss). let (k,s) = class_to_term_intern j t
              in (k,s # ss)) ts (i,[]) in (k, Fun f ss))"
     |  "class_to_term_intern i (Var (Inl _)) = (i+1,Var (Inl i))"
     |  "class_to_term_intern i (Var (Inr x)) = (i,Var (Inr x))"

definition class_to_term :: "('f,unit + 'v)term \<Rightarrow> ('f,int + 'v)term"
  where "class_to_term t \<equiv> snd (class_to_term_intern 0 t)"

lemma class_to_term_intern_vars: assumes "class_to_term_intern i t = (j,s)"
  shows "i \<le> j \<and> (Inl k \<in> vars_term s \<longrightarrow> i \<le> k \<and> k < j)"
using assms
proof (induct t arbitrary: i j s)
  case (Var x i j s)
  thus ?case by (cases x, auto)
next
  case (Fun f ts i j sss)
  from Fun(2) obtain ss where sss: "sss = Fun f ss"
    by (cases sss, auto split: prod.split_asm)
  note Fun = Fun[unfolded sss]
  let ?f = "(\<lambda> t (j,ss). let (k,s) = class_to_term_intern j t
              in (k,s # ss))"
  from Fun(2) have id: "foldr ?f ts (i,[]) = (j,ss)" by (auto split: prod.split_asm)
  with Fun(1) have "i \<le> j \<and> (\<forall> s \<in> set ss. Inl k \<in> vars_term s \<longrightarrow> i \<le> k \<and> k < j)"
  proof (induct ts arbitrary: i j ss)
    case Nil thus ?case by auto
  next
    case (Cons t ts i j sss)
    obtain i' ss where ts: "foldr ?f ts (i,[]) = (i',ss)" by (cases "foldr ?f ts (i,[])", auto)
    obtain j' s where t: "class_to_term_intern i' t = (j',s)" by (cases "class_to_term_intern i' t", auto)
    from ts t Cons(3) have sss: "sss = s # ss" and j: "j = j'" by auto
    from Cons(2)[OF _ t] have t: "i' \<le> j' \<and> (Inl k \<in> vars_term s \<longrightarrow> i' \<le> k \<and> k < j')" by auto 
    have ts: "i \<le> i' \<and> (\<forall> s \<in> set ss. Inl k \<in> vars_term s \<longrightarrow> i \<le> k \<and> k < i')"
      by (rule Cons(1)[OF _ ts], insert Cons(2), auto)
    from t ts j sss show ?case by auto
  qed
  thus ?case unfolding sss by auto
qed

lemma class_to_term_intern: 
  "[t]_\<sigma> = {snd (class_to_term_intern i t) \<cdot> (merge_substs \<tau> \<sigma>)| \<tau>. True}" 
proof (induct t arbitrary: i)
  case (Var x i)
  show ?case 
    by (cases x, auto)
next
  case (Fun f ts i)
  show ?case (is "?l (Fun f ts) = ?r i (Fun f ts)")
  proof -
    let ?f = "(\<lambda> t (i,ss). let (j,s) = class_to_term_intern i t
      in (j,s # ss))"
    let ?g = "(\<lambda> t (i,ss,is,js). let (j,s) = class_to_term_intern i t
      in (j,s # ss, i # is, j # js))"
    obtain j us ii jj where foldg: "foldr ?g ts (i,[],[],[]) = (j, us, ii, jj)" by (cases "foldr ?g ts (i, [], [], [])", auto)
    let ?P = "\<lambda> ts us ii jj k j. class_to_term_intern (ii ! k) (ts ! k)  = (jj ! k, us ! k) \<and> jj ! k \<le> j \<and> (\<forall> k'. k < k' \<and> k' < length ts \<longrightarrow> jj ! k' \<le> ii ! k)"
    from foldg have fold_prop: "i \<le> j \<and> foldr ?f ts (i,[]) = (j, us) \<and> length us = length ts \<and> length ii = length ts \<and> length jj = length ts \<and> (\<forall> k < length ts. ?P ts us ii jj k j)"
    proof (induct ts arbitrary: j us ii jj)
      case Nil thus ?case by simp
    next
      case (Cons t ts k us' ii' jj')
      obtain i' us ii jj where foldg: "foldr ?g ts (i,[],[],[]) = (i', us, ii, jj)" by (cases "foldr ?g ts (i, [], [], [])", auto)
      obtain j' u where cjt: "class_to_term_intern i' t = (j',u)" by (cases "class_to_term_intern i' t", auto)
      from foldg cjt have fold: "foldr ?g (t # ts) (i,[],[],[]) = (j',u # us, i' # ii, j' # jj)" by auto
      from fold[unfolded Cons(2)] have k: "k = j'" and us: "us' = u # us" and ii: "ii' = i' # ii" and jj: "jj' = j' # jj" by auto
      from Cons(1)[OF foldg] have ii': "i \<le> i'" and 
        foldf: "foldr ?f ts (i,[]) = (i', us)" and 
        lus: "length us = length ts" and 
        lii: "length ii = length ts" and 
        ljj: "length jj = length ts" and 
        comp: "\<forall>k < length ts. ?P ts us ii jj k i'"
        by auto
      from class_to_term_intern_vars[OF cjt] have i'j': "i' \<le> j'" by simp
      from ii' i'j'  have ij': "i \<le> j'" by auto
      show ?case
      proof (intro conjI,
          simp add: k ij', 
          simp add: foldf[simplified] us k cjt,
          simp add: us lus,
          simp add: ii lii,
          simp add: jj ljj, 
          intro allI impI, unfold k)
        fix k
        assume k: "k < length (t # ts)"
        show "?P (t # ts) us' ii' jj' k j'"
        proof (cases k)
          case 0
          show ?thesis unfolding us ii jj 0
          proof (simp add: cjt, intro allI impI)
            fix k'
            assume "0 < k' \<and> k' < Suc (length ts)"
            then obtain k'' where k': "k' = Suc k''" and len: "k'' < length ts" by (cases k', auto)
            from comp[THEN spec, THEN mp[OF _ len]] have cik: "class_to_term_intern (ii ! k'') (ts ! k'') = (jj ! k'', us ! k'')" 
              and ji': "jj ! k'' \<le> i'" by auto
            thus "(j' # jj) ! k' \<le> i'" unfolding k' by simp
          qed
        next
          case (Suc k)
          with k have k: "k < length ts" by auto
          from comp[THEN spec, THEN mp[OF _ k]] have cik: "class_to_term_intern (ii ! k) (ts ! k) = (jj ! k, us ! k)" 
            and ji': "jj ! k \<le> i'" by auto
          from class_to_term_intern_vars[OF cik] have ij: "ii ! k \<le> jj ! k" by simp
          from ji' i'j' have jjk: "jj ! k \<le> j'" by simp
          show ?thesis unfolding us ii jj Suc
          proof (intro conjI,
              simp add: cik,
              simp add: jjk,
              intro allI impI)
            fix k'
            assume ass: "Suc k < k' \<and> k' < length (t # ts)"
            then obtain k'' where k': "k' = Suc k''" and kk: "k < k'' \<and> k'' < length ts" by (cases k', auto)
            from comp[THEN spec, THEN mp[OF _ k], THEN conjunct2, THEN conjunct2, THEN spec, THEN mp[OF _ kk]] 
            show "(j' # jj) ! k' \<le> (i' # ii) ! Suc k" unfolding k' by simp
          qed
        qed
      qed
    qed
    from fold_prop have ct: "\<And> k. k < length ts \<Longrightarrow> ?P ts us ii jj k j" by auto
    from fold_prop have id: "snd (class_to_term_intern i (Fun f ts)) = (Fun f us)"
      by (simp add: Let_def)          
    {
      fix s
      assume "s \<in> ?l (Fun f ts)"      
      then obtain ss where len: "length ss = length ts" and ss: "\<And> i. i < length ts \<Longrightarrow> ss ! i \<in> [ts ! i]_\<sigma>" and s: "s = Fun f ss" by auto
      {
        fix i j
        assume i: "i < length ts"
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] ss[OF i] have "ss ! i  \<in> ?r j (ts ! i)" by auto
      } note ss = this
      from fold_prop len have len2: "length ss = length us" by auto
      {
        fix k
        assume k: "k < length ts"
        from ct[OF k] ss[OF k, of "ii ! k"]
        have "\<exists> \<tau>. ss ! k = us ! k \<cdot> merge_substs \<tau> \<sigma>"
          by auto
      } note arg = this
      from arg have "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> ss ! i = us ! i \<cdot> merge_substs \<tau> \<sigma>" by auto
      from choice[OF this] obtain \<tau>i where arg: "\<And> i. i < length ts \<Longrightarrow> ss ! i = us ! i \<cdot> merge_substs (\<tau>i i) \<sigma>" by auto
      let ?choice = "\<lambda> x. (THE i. i < length ts \<and> Inl x \<in> vars_term (us ! i))"
      let ?\<tau> = "\<lambda> x. \<tau>i (?choice x) x"
      have "s \<in> ?r i (Fun f ts)" unfolding s id
      proof (simp, rule exI, rule sym, unfold map_nth_eq_conv[OF len2[symmetric]], clarify)
        fix i
        assume "i < length ss"
        with len len2 have i: "i < length ts" by auto
        have "ss ! i = us ! i \<cdot> merge_substs ?\<tau> \<sigma>"
          unfolding arg[OF i] term_subst_eq_conv
        proof (intro ballI)
          fix x
          assume x: "x \<in> vars_term (us ! i)"
          show "merge_substs (\<tau>i i) \<sigma> x = merge_substs ?\<tau> \<sigma> x"
          proof (cases x)
            case (Inr y)
            show ?thesis unfolding Inr  by simp
          next
            case (Inl y)
            have ci: "?choice y = i"
            proof
              show "i < length ts \<and> Inl y \<in> vars_term (us ! i)" using x i Inl  by auto
            next
              fix j
              assume "j < length ts \<and> Inl y \<in> vars_term (us ! j)"
              hence j: "j < length ts" and y: "Inl y \<in> vars_term (us ! j)" by auto
              from class_to_term_intern_vars[OF ct[OF i, THEN conjunct1], THEN conjunct2, THEN mp[OF _ x[unfolded Inl]]]
              have iy: "ii ! i \<le> y" and yi: "y < jj ! i" by auto
              from class_to_term_intern_vars[OF ct[OF j, THEN conjunct1], THEN conjunct2, THEN mp[OF _ y]]
              have jy: "ii ! j \<le> y" and yj: "y < jj ! j" by auto
              from iy yj have ij: "ii ! i < jj ! j" by auto
              from jy yi have ji: "ii ! j < jj ! i" by auto
              {
                fix i j
                assume i: "i < length ts" and j: "j < length ts" and gt: "ii ! i < jj ! j"
                have "j \<le> i"
                proof (rule ccontr)
                  assume "\<not> j \<le> i"
                  hence "i < j" by auto
                  with ct[OF i, THEN conjunct2, THEN conjunct2, THEN spec[of _ j]]
                    j have "jj ! j \<le> ii ! i" by auto
                  with gt show False by auto
                qed
              } note increasing = this
              from increasing[OF i j ij] increasing[OF j i ji]
              show "j = i" by auto
            qed
            show ?thesis unfolding Inl 
              by (simp add: ci)
          qed
        qed
        thus "us ! i \<cdot> sum_case ?\<tau> \<sigma> = ss ! i" by simp
      qed                          
    }
    moreover
    {
      fix s
      assume "s \<in> ?r i (Fun f ts)"
      with fold_prop[THEN conjunct2, THEN conjunct1]      
      obtain \<tau> where s: "s = Fun f (map (\<lambda> t. t \<cdot> merge_substs \<tau> \<sigma>) us)" (is "_ = Fun f (map ?sig us)") by auto
      from fold_prop have len: "length ts = length us" by auto
      let ?P = "\<lambda> i tsi. ?sig (us ! i) = tsi \<and> tsi \<in> [ts ! i]_\<sigma>"
      {
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by auto
        with ct[OF i] have us: "us ! i = snd (class_to_term_intern (ii ! i) (ts ! i))" by auto
        from Fun[OF mem]
        have "?sig (us ! i) \<in> [ts ! i]_\<sigma>" unfolding us by auto
        hence "\<exists> tsi. ?P i tsi" by auto
      } 
      hence "\<forall> i. \<exists> tsi. i < length ts \<longrightarrow> ?P i tsi" by auto
      from choice[OF this] obtain ss where ss: "\<And> i. i < length ts \<Longrightarrow> ?P i (ss i)" by auto      
      let ?ss = "map ss [0 ..< length ts]"
      have "s \<in> ?l (Fun f ts)" unfolding s using ss by (simp add: len ss)
    }
    ultimately show ?case by blast
  qed
qed

lemma class_to_term: "[t]_\<sigma> = {class_to_term t \<cdot> (merge_substs \<tau> \<sigma>)| \<tau>. True}" 
  unfolding class_to_term_def 
  by (rule class_to_term_intern)


(* usually P1 and P2 are checks on whether a term is in normal form w.r.t. Q *)
(* I made this generic since perhaps one will also need other forms of check *)
definition check_class  :: "(('f, 'v, 'v + int + 'w)gsubst \<Rightarrow> bool) \<Rightarrow> 
      (('f, int + 'w, 'v + int + 'w)gsubst \<Rightarrow> bool) \<Rightarrow> 
      ('f,'v)term \<Rightarrow> ('f,unit + 'w)term \<Rightarrow> bool"
  where "check_class P1 P2 s t \<equiv> case mgu_var_disjoint s (class_to_term t) of  None \<Rightarrow> False | Some (\<mu>1,\<mu>2) \<Rightarrow> P1 \<mu>1 \<and> P2 \<mu>2"



lemma check_class_complete: fixes s :: "('f,'v)term" and t :: "('f,unit+'w)term" 
  and S :: "('f,'v)terms" and T :: "('f,'w)terms" 
 and Pmu2 :: "('f, int + 'w, 'v + int + 'w)gsubst \<Rightarrow> bool" 
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. Pm (u \<cdot> \<mu>))"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. Pm (u \<cdot> map_subst_domain Inr \<mu>))" 
   and P: "\<And> u. P u \<Longrightarrow> Pm (map_vars Inl u)" 
   and Pm: "\<And> u \<gamma>. Pm (u \<cdot> \<gamma>) \<Longrightarrow> Pm u"
   and  R: "(\<exists> \<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall> u \<in> S. P (u \<cdot> \<sigma>)) \<and> (\<forall> u \<in> T. P (u \<cdot> \<tau>)))" (is ?R)
  shows "(check_class Pmu1 Pmu2 s t)" (is ?L)
proof -
  from R obtain \<sigma> \<tau> where mem: "s \<cdot> \<sigma> \<in> [t]_\<tau>"
  and S: "\<And> u. u \<in> S \<Longrightarrow> P (u \<cdot> \<sigma>)"
  and T: "\<And> u. u \<in> T \<Longrightarrow> P (u \<cdot> \<tau>)" by auto
  from mem[unfolded class_to_term]
  obtain \<tau>' where unify: "s \<cdot> \<sigma> = class_to_term t \<cdot> merge_substs \<tau>' \<tau>" by auto
  obtain \<tau>'' where  \<tau>'': "\<tau>'' = merge_substs \<tau>' \<tau>" by auto
  obtain \<gamma> where \<gamma>: "\<gamma> = merge_substs \<sigma> \<tau>''" by auto
  obtain ct where ct: "ct = class_to_term t" by auto
  let ?map = "map_vars_subst_ran (Inl :: 'v \<Rightarrow> 'v + int + 'w)"
  let ?sig = "?map \<sigma>"
  let ?tau = "?map \<tau>''"
  from unify have "s \<cdot> \<sigma> = ct \<cdot> \<tau>''" unfolding \<tau>'' ct by auto
  hence "(map_vars Inl (s \<cdot> \<sigma>) :: ('f,'v + int + 'w)term) = map_vars Inl (ct \<cdot> \<tau>'')" by auto
  (* perform type conversion for mgu_var_disjoint *)
  hence "s \<cdot> ?sig = ct \<cdot> ?tau" unfolding map_vars_subst_ran .
  from mgu_var_disjoint_instance[OF this] obtain \<mu>1 \<mu>2 \<delta>1 \<delta>2
    where mu: "mgu_var_disjoint s ct = Some (\<mu>1, \<mu>2)" and
    sig: "?sig = \<mu>1 \<circ>s \<delta>1" and tau: "?tau = \<mu>2 \<circ>s \<delta>2" by auto
  have pmu1: "Pmu1 \<mu>1" unfolding Pmu1
  proof (intro ballI conjI)
    fix u
    assume u: "u \<in> S"
    from P[OF S[OF u]] have "Pm (map_vars Inl (u \<cdot> \<sigma>))" by auto
    from this[unfolded map_vars_subst_ran sig]
    have "Pm (u \<cdot> \<mu>1 \<cdot> \<delta>1)" by auto
    thus "Pm (u \<cdot> \<mu>1)" by (rule Pm)
  qed
  have pmu2: "Pmu2 \<mu>2" unfolding Pmu2
  proof (intro ballI conjI)
    fix u
    assume u: "u \<in> T"
    from P[OF T[OF u]] have "Pm (map_vars Inl (u \<cdot> \<tau>))" by auto
    hence "Pm (map_vars Inr u \<cdot> ?tau)" 
      unfolding \<tau>'' map_vars_subst_ran[symmetric]  merge_substs_right .
    from this[unfolded tau]
    have "Pm (map_vars Inr u \<cdot> \<mu>2 \<cdot> \<delta>2)" by auto
    from Pm[OF this]
    show "Pm (u \<cdot> map_subst_domain Inr \<mu>2)" unfolding map_subst_domain .
  qed
  show ?L unfolding check_class_def mu[unfolded ct] 
    by (simp add: pmu1 pmu2)
qed

lemma check_class_sound: fixes s :: "('f,'v)term" and t :: "('f,unit+'w)term" 
  and S :: "('f,'v)terms" and T :: "('f,'w)terms" 
 and Pmu2 :: "('f, int + 'w, 'v + int + 'w)gsubst \<Rightarrow> bool" 
  and P :: "('f,'v)term \<Rightarrow> bool"
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. Pm (u \<cdot> \<mu>))"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. Pm (u \<cdot> map_subst_domain Inr \<mu>))" 
   and P: "\<And> u. Pm u \<Longrightarrow> P (map_vars conv u)" 
   and  L: "(check_class Pmu1 Pmu2 s t)" (is ?L)
  shows "\<exists> \<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall> u \<in> S. P (u \<cdot> \<sigma>)) \<and> (\<forall> u \<in> T. P (u \<cdot> \<tau>))" (is ?R)
proof -
  obtain ct where ct: "ct = class_to_term t" by auto
  from L[unfolded check_class_def]
  obtain \<mu>1 \<mu>2 where mu: "mgu_var_disjoint s ct = Some (\<mu>1,\<mu>2)" and pmu1: "Pmu1 \<mu>1" and pmu2: "Pmu2 \<mu>2" unfolding ct[symmetric]
    by (cases "mgu_var_disjoint s ct", auto)
  from mgu_var_disjoint_sound[OF mu]
  have id: "s \<cdot> \<mu>1 =  ct \<cdot> \<mu>2" .
  from pmu1[unfolded Pmu1] 
  have S: "\<And> u. u \<in> S \<Longrightarrow> Pm (u \<cdot> \<mu>1)" by auto
  from pmu2[unfolded Pmu2]
  have  T: "\<And> u. u \<in> T \<Longrightarrow> Pm (u \<cdot> map_subst_domain Inr \<mu>2)" by auto
  let ?sig = "map_vars_subst_ran conv \<mu>1" 
  let ?tau = "map_subst_domain Inr (map_vars_subst_ran conv \<mu>2)"
  obtain \<tau>' where \<tau>': "\<tau>' = (\<lambda>y. map_vars conv (\<mu>2 (Inl y)))" by auto
  have \<tau>': "merge_substs \<tau>' ?tau = map_vars_subst_ran conv \<mu>2" (is "?l = ?r")
  proof
    fix x
    show "?l x = ?r x"
      by (cases x, auto simp: \<tau>')
  qed
  have mem: "s \<cdot> ?sig \<in> [t]_?tau"
    unfolding class_to_term ct[symmetric]
    by (rule, rule exI[of _ \<tau>'], unfold \<tau>', unfold map_vars_subst_ran[symmetric] id, simp)
  show ?R 
  proof (intro conjI exI ballI, rule mem)
    fix u
    assume "u \<in> S"
    from S[OF this]
    show "P (u \<cdot> ?sig)" unfolding map_vars_subst_ran[symmetric]
      by (rule P)
  next
    fix u
    assume "u \<in> T"
    from T[OF this]
    show "P (u \<cdot> ?tau)" unfolding map_subst_domain[symmetric] map_vars_subst_ran[symmetric]
      by (rule P)
  qed
qed

lemma check_class_nf_complete: fixes s :: "('f,'v)term" and t :: "('f,unit+'w)term" 
  and S :: "('f,'v)terms" and T :: "('f,'w)terms" 
  and Q :: "('f,'v)terms"
 and Pmu2 :: "('f, int + 'w, 'v + int + 'w)gsubst \<Rightarrow> bool" 
  defines QQ: "QQ \<equiv> map_vars Inl ` Q" 
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. u \<cdot> \<mu> \<in> NF_terms QQ)"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. u \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms QQ)"
   and  R: "(\<exists> \<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall> u \<in> S. u \<cdot> \<sigma> \<in> NF_terms Q) \<and> (\<forall> u \<in> T. u \<cdot> \<tau> \<in> NF_terms Q))"
  shows "(check_class Pmu1 Pmu2 s t)"
proof (rule check_class_complete[OF Pmu1 Pmu2 _ _ R])
  fix u :: "('f,'v + int + 'w)term" and \<sigma>
  assume "u \<cdot> \<sigma> \<in> NF_terms QQ" thus "u \<in> NF_terms QQ" unfolding NF_def by auto
next
  fix u  
  assume u: "u \<in> NF_terms Q" 
  show "map_vars Inl u \<in> NF_terms QQ" unfolding QQ
    by (rule NF_map_vars_inj[OF _ u, of "Sum_Type.Projl"], auto)
qed


lemma inj_map_vars_the_inv:
  assumes "inj f"
  shows "map_vars (the_inv f) (map_vars f t) = t"
  unfolding map_vars_compose o_def the_inv_f_f[OF assms]
  by simp

lemma map_vars_Projl_Inl[simp]:
  "map_vars Sum_Type.Projl (map_vars Inl t) = t"
  unfolding map_vars_compose o_def by auto

lemma map_vars_NF_terms:
  assumes inj: "inj f"
    and NF: "t \<in> NF_terms (map_vars Inl ` T)" (is "t \<in> NF_terms ?T")
  shows "map_vars f t \<in> NF_terms T" (is "?t \<in> NF_terms T")
proof (rule ccontr)
  assume "?t \<notin> NF_terms T"
  then obtain u where "(?t, u) \<in> rstep (Id_on T)" by (auto simp: NF_def)
  then obtain C \<sigma> l r where in_T: "(l, r) \<in> Id_on T"
    and t': "?t = C\<langle>l \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by (rule rstepE)
  let ?f = "the_inv f"
  let ?C = "map_vars_ctxt ?f C"
  let ?l = "map_vars Inl l"
  let ?r = "map_vars Inl r"
  let ?\<tau> = "map_vars_subst_ran ?f \<sigma>"
  let ?\<sigma> = "map_subst_domain Sum_Type.Projl (map_vars_subst_ran (the_inv f) \<sigma>)"
  from in_T have "(?l, ?r) \<in> Id_on ?T" by auto
  moreover have "t = ?C\<langle>?l \<cdot> ?\<sigma>\<rangle>"
  proof -
    have "map_vars ?f ?t = map_vars ?f (C\<langle>l \<cdot> \<sigma>\<rangle>)" by (simp add: t')
    hence "t = map_vars ?f (C\<langle>l \<cdot> \<sigma>\<rangle>)" by (simp add: inj_map_vars_the_inv[OF inj])
    also have "\<dots> = ?C\<langle>map_vars ?f (l \<cdot> \<sigma>)\<rangle>" by (simp add: map_vars_ctxt_commute)
    also have "\<dots> = ?C\<langle>l \<cdot> ?\<tau>\<rangle>" by (simp add: map_vars_subst_ran)
    also have "\<dots> = ?C\<langle>map_vars Sum_Type.Projl ?l \<cdot> ?\<tau>\<rangle>" by simp
    finally show ?thesis unfolding map_subst_domain .
  qed
  moreover have "map_vars ?f u = ?C\<langle>?r \<cdot> ?\<sigma>\<rangle>"
  proof -
    have "map_vars ?f u = ?C\<langle>map_vars ?f (r \<cdot> \<sigma>)\<rangle>" by (simp add: u map_vars_ctxt_commute)
    also have "\<dots> = ?C\<langle>?r \<cdot> ?\<sigma>\<rangle>" unfolding map_subst_domain[symmetric] map_vars_subst_ran[symmetric] by simp
    finally show ?thesis .
  qed
  ultimately have "(t, map_vars ?f u) \<in> rstep (Id_on ?T)" by (rule rstepI)
  with NF show False by auto
qed

lemma check_class_nf_sound:
  fixes s :: "('f,'v)term" and t :: "('f, unit + 'w) term" 
    and S :: "('f,'v)terms" and T :: "('f, 'w) terms" 
    and Q :: "('f,'v)terms"
    and Pmu2 :: "('f, int + 'w, 'v + int + 'w) gsubst \<Rightarrow> bool" 
  defines QQ: "QQ \<equiv> map_vars Inl ` Q" 
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. u \<cdot> \<mu> \<in> NF_terms QQ)"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. u \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms QQ)"
   and conv: "\<exists>conv::('v + int + 'w \<Rightarrow> 'v). inj conv"
   and  L: "(check_class Pmu1 Pmu2 s t)" (is ?L)
  shows "\<exists>\<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall>u \<in> S. u \<cdot> \<sigma> \<in> NF_terms Q) \<and> (\<forall> u \<in> T. u \<cdot> \<tau> \<in> NF_terms Q)"
    (is ?R)
proof -
  from conv obtain conv :: "'v + int + 'w \<Rightarrow> 'v" where "inj conv" by auto
  {
    fix u
    assume u: "u \<in> NF_terms QQ"
    from map_vars_NF_terms[OF `inj conv` this[unfolded QQ]]
      have "map_vars conv u \<in> NF_terms Q" .
  }
  thus ?thesis
    by (rule check_class_sound[OF Pmu1 Pmu2 _ L]) simp
qed  
  

lemma check_class_nf: fixes s :: "('f,'v :: countable)term" and t :: "('f,unit+('w :: countable))term" 
  and S :: "('f,'v)terms" and T :: "('f,'w)terms" 
  and Q :: "('f,'v)terms"
 and Pmu2 :: "('f, int + 'w, 'v + int + 'w)gsubst \<Rightarrow> bool" 
  defines QQ: "QQ \<equiv> map_vars Inl ` Q" 
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. u \<cdot> \<mu> \<in> NF_terms QQ)"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. u \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms QQ)"
   and inf: "\<not> finite (UNIV :: 'v set)"
  shows "(check_class Pmu1 Pmu2 s t) = (\<exists> \<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall> u \<in> S. u \<cdot> \<sigma> \<in> NF_terms Q) \<and> (\<forall> u \<in> T. u \<cdot> \<tau> \<in> NF_terms Q))" (is "?L = ?R")
proof
  assume L: ?L
  from ex_inj obtain g :: "'v + int + 'w \<Rightarrow> nat" where g: "inj g" ..
  from infinite_countable_subset[OF inf] obtain h :: "nat \<Rightarrow> 'v" where h: "inj h" by auto
  from inj_comp[OF h g] have "inj (h \<circ> g)" .
  then obtain f :: "'v + int + 'w \<Rightarrow> 'v" where f: "inj f" by auto
  show ?R
    by (rule check_class_nf_sound[OF Pmu1[unfolded QQ] Pmu2[unfolded QQ] _ L], insert f, auto)
next
  assume R: ?R
  show ?L
    by (rule check_class_nf_complete[OF Pmu1[unfolded QQ] Pmu2[unfolded QQ] R])
qed

lemma check_class_nf_string: fixes s :: "('f,string)term" and t :: "('f,unit+string)term" 
  and S :: "('f,string)terms" and T :: "('f,string)terms" 
  and Q :: "('f,string)terms"
 and Pmu2 :: "('f, int + string, string + int + string)gsubst \<Rightarrow> bool" 
  defines QQ: "QQ \<equiv> map_vars Inl ` Q" 
  assumes Pmu1: "\<And> \<mu>. Pmu1 \<mu> = (\<forall> u \<in> S. u \<cdot> \<mu> \<in> NF_terms QQ)"
   and Pmu2: "\<And> \<mu>. Pmu2 \<mu> = (\<forall> u \<in> T. u \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms QQ)"
  shows "(check_class Pmu1 Pmu2 s t) = (\<exists> \<sigma> \<tau>. s \<cdot> \<sigma> \<in> [t]_\<tau> \<and> (\<forall> u \<in> S. u \<cdot> \<sigma> \<in> NF_terms Q) \<and> (\<forall> u \<in> T. u \<cdot> \<tau> \<in> NF_terms Q))" (is "?L = ?R")
  by (rule check_class_nf[OF Pmu1[unfolded QQ] Pmu2[unfolded QQ] infinite_UNIV_listI])

(*FIXME: move*)
abbreviation NF_trs :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set" where
  "NF_trs R \<equiv> NF (rstep R)"

fun
  icap ::
    "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, unit + 'v) term"
where
  "icap R Q S (Var x) =
    Var (if NF_terms Q \<subseteq> NF_trs R \<and> (\<exists>t \<in> S. x \<in> vars_term t) then Inr x else Inl ())"
| "icap R Q S (Fun f ts) = (
    let t' = Fun f (map (icap R Q S) ts) in
    if (\<exists>r \<in> R. check_class 
         (\<lambda> \<mu>. (\<forall> u \<in> set (args (fst r)). u \<cdot> \<mu> \<in> NF_terms (map_vars Inl ` Q)))
         (\<lambda> \<mu>. (\<forall> u \<in> S. u \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms (map_vars Inl ` Q)))
         (fst r) t')
      then Var (Inl ())
      else t')"

lemma icap: "is_ecap icap"
proof(rule, intro conjI)
  fix R :: "('f,'v)trs" and Q :: "('f,'v)terms" and S t f ts
  show "icap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (icap R Q S) ts)}"
    by (auto simp: Let_def)
next
  fix R :: "('f,'v)trs" and Q :: "('f,'v)terms" and S t f ts
  {
    fix \<tau> :: "('f,'v)subst"
    have "[cap R Q S t]_\<tau> \<subseteq> [icap R Q S t]_\<tau>"
    proof (induct t arbitrary: \<tau>)
      case (Var x)
      show ?case 
      proof (cases "\<exists>\<sigma>. S \<cdot>set \<sigma> \<subseteq> NF_terms Q \<and> Var x \<cdot> \<sigma> \<notin> NF (qrstep Q R)")
        case False
        thus ?thesis by auto
      next
        case True
        then obtain \<sigma> where subset: "S \<cdot>set \<sigma> \<subseteq> NF_trs (Id_on Q)" and notNF: "Var x \<cdot> \<sigma> \<notin> NF (qrstep Q R)" by auto
        show ?thesis 
        proof (cases "NF_trs (Id_on Q) \<subseteq> NF_trs R \<and> (\<exists>t \<in> S. x \<in> vars_term t)")
          case False thus ?thesis by auto
        next
          case True
          then obtain t where t: "t \<in> S" and x: "x \<in> vars_term t" and subset2: "NF_trs (Id_on Q) \<subseteq> NF_trs R" by auto
          from t subset subset2 have NF: "t \<cdot> \<sigma> \<in> NF_trs R" by blast
          from x have "Var x \<unlhd> t" by auto
          hence "Var x \<cdot> \<sigma> \<unlhd> t \<cdot> \<sigma>" by auto
          from NF_subterm[OF NF this] have "Var x \<cdot> \<sigma> \<in> NF_trs R" by simp
          with notNF have False unfolding NF_def using qrstep_subset_rstep[of Q R] by auto
          thus ?thesis by auto
        qed
      qed
    next
      case (Fun f ss)
      let ?Pm = "\<lambda> t. t \<in> NF_terms (map_vars (Inl :: ('v \<Rightarrow> 'v + int + 'v)) ` Q)"
      let ?Pl = "\<lambda> r \<mu>. (\<forall> u \<in> set (args (fst r)). ?Pm (u \<cdot> \<mu>))"
      let ?Pr = "\<lambda> r \<mu>. (\<forall> u \<in> S. ?Pm  (u \<cdot> map_subst_domain Inr \<mu>))"
      show ?case
      proof (cases "\<exists> r \<in> R. check_class (?Pl r) (?Pr r) (fst r) (Fun f (map (icap R Q S) ss))")
        case True
        thus ?thesis by (auto simp: Let_def)
      next
        case False 
        hence icap: "icap R Q S (Fun f ss) = Fun f (map (icap R Q S) ss)" by (auto simp: Let_def)
        {
          fix \<tau> :: "('f,'v)subst"
          from Fun[of _ \<tau>, unfolded set_conv_nth] 
          have "[Fun f (map (cap R Q S) ss)]_\<tau> \<subseteq> [Fun f (map (icap R Q S) ss)]_\<tau>" by auto
        } note ind = this
        show ?thesis       
        proof (cases "\<exists> \<sigma>. S \<cdot>set \<sigma> \<subseteq> NF_trs (Id_on Q) \<and> (\<exists> u. (Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O rqrstep Q R)")
          case False
          hence cap: "cap R Q S (Fun f ss) = Fun f (map (cap R Q S) ss)" by auto
          show ?thesis unfolding cap icap by (rule ind)
        next
          case True
          then obtain \<sigma> u where S: "S \<cdot>set \<sigma> \<subseteq> NF_trs (Id_on Q)" and steps: "(Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O rqrstep Q R" by auto
          from first_step_O[OF _ steps, unfolded qrstep_iff_rqrstep_or_nrqrstep, of "nrqrstep Q R"]
          obtain u where "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrqrstep Q R)^* O rqrstep Q R" 
            by auto
          then obtain v where steps: "(Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss), v) \<in> (nrqrstep Q R)^*" and step: "(v,u) \<in> rqrstep Q R" by auto
          from nrqrsteps_preserve_root[OF steps] obtain ts where v: "v = Fun f ts" by auto
          from nrqrsteps_num_args[OF steps] v have len: "length ts = length ss" by auto
          {
            fix i
            assume i: "i < length ss"
            hence si: "ss ! i \<in> set ss" by auto
            from nrqrsteps_imp_arg_qrsteps[OF steps, unfolded v, of i]
            have "(ss ! i \<cdot> \<sigma>, ts ! i) \<in> (qrstep Q R)^*" using len i by auto
            from ecap_steps[OF is_ecap_cap this S] have "ts ! i \<in> [cap R Q S (ss ! i)]_\<sigma>" by auto
            with Fun[OF si, of \<sigma>] have "ts ! i \<in> [icap R Q S (ss ! i)]_\<sigma>" by auto
          } note icap_arg = this
          have in_icap: "v \<in> [Fun f (map (icap R Q S) ss)]_\<sigma>" unfolding v 
            by (simp add: len icap_arg)
          from step[unfolded rqrstep_def] obtain l r \<tau> where v: "v = l \<cdot> \<tau>" and lr: "(l,r) \<in> R" 
            and NF: "\<And> u. u \<lhd> l \<cdot> \<tau> \<Longrightarrow> u \<in> NF_terms Q" by auto
          from False lr have ncheck: "\<not> check_class (?Pl (l,r)) (?Pr (l,r)) l (Fun f (map (icap R Q S) ss))" by auto
          have check: "check_class (?Pl (l,r)) (?Pr (l,r)) l (Fun f (map (icap R Q S) ss))" 
          proof (rule check_class_nf_complete[of _ "set (args l)" Q _ S], simp, simp,
            intro exI conjI, rule in_icap[unfolded v])
            from S show "\<forall> u \<in> S. u \<cdot> \<sigma> \<in> NF_terms Q" by auto
          next
            show "\<forall> u \<in> set (args l). u \<cdot> \<tau> \<in> NF_terms Q"
            proof
              fix u
              assume u: "u \<in> set (args l)" 
              from u obtain f ls where "l = Fun f ls" and "u \<in> set ls" by (cases l, auto)
              hence "u \<cdot> \<tau> \<lhd> l \<cdot> \<tau>" by auto
              from NF[OF this] show "u \<cdot> \<tau> \<in> NF_terms Q" .
            qed
          qed
          with ncheck
          show ?thesis ..
        qed
      qed
    qed
  }
  thus "[cap R Q S t] \<subseteq> [icap R Q S t]" by auto
qed          

definition
  IEDG :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) trs \<Rightarrow> (('f, 'v)rule \<times> ('f,'v)rule) set"
where
  "IEDG P Q R \<equiv> {((s,t), (u,v)). check_class 
         (\<lambda> \<mu>. u \<cdot> \<mu> \<in> NF_terms (map_vars Inl ` Q))
         (\<lambda> \<mu>. s \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms (map_vars Inl ` Q))
         u (icap R Q {s} t)}"

lemma IEDG: "DG P Q R \<subseteq> IEDG P Q R"
proof -
  {
    fix s t u v
    assume "((s,t),(u,v)) \<in> DG P Q R"
    from this[unfolded DG_def] obtain \<sigma> \<tau> where 
      steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep Q R)^*"
      and s: "s \<cdot> \<sigma> \<in> NF_terms Q" and u: "u \<cdot> \<tau> \<in> NF_terms Q"
      by auto
    from ecap_steps[OF icap steps, of "{s}"] s 
    have cap: "u \<cdot> \<tau> \<in> [icap R Q {s} t]_\<sigma>" by auto
    let ?cl = "(\<lambda> \<mu>. u \<cdot> \<mu> \<in> NF_terms (map_vars Inl ` Q))"
    let ?cr = "(\<lambda> \<mu>. s \<cdot> map_subst_domain Inr \<mu> \<in> NF_terms (map_vars Inl ` Q))"
    have "check_class ?cl ?cr u (icap R Q {s} t)" 
      by (rule check_class_nf_complete[of _ "{u}" Q _ "{s}"], simp, simp, intro exI conjI, rule cap, simp add: u, simp add: s)
    hence "((s,t),(u,v)) \<in> IEDG P Q R"
      unfolding IEDG_def by auto
  }
  thus ?thesis by auto
qed


end



