(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Generic Error Monads *}

theory Error_Monad
imports "~~/src/HOL/Library/Monad_Syntax"
begin


text {*
The \emph{monadic bind} operation.
*}
definition 
  bind :: "'e + 'a \<Rightarrow> ('a \<Rightarrow> 'e + 'b) \<Rightarrow> 'e + 'b"
where
  "bind m f \<equiv> case m of Inr x \<Rightarrow> f x | e \<Rightarrow> e"

setup {*
  Adhoc_Overloading.add_variant @{const_name Monad_Syntax.bind} @{const_name bind}
*}

abbreviation (input) return where "return \<equiv> Inr"
abbreviation (input) error where "error \<equiv> Inl"
abbreviation (input) run where "run \<equiv> Sum_Type.Projr"


subsection {* Monad Laws *}

lemma return_bind: "(Inr x \<guillemotright>= f) = f x" by (simp add: bind_def)

lemma bind_return: "(m \<guillemotright>= Inr) = m" by (cases m) (simp_all add: bind_def)

lemma error_bind: "(error e \<guillemotright>= f) = error e" by (simp add: bind_def)

lemma bind_assoc: "(((m :: 'a + 'b) \<guillemotright>= f) \<guillemotright>= g) = (m \<guillemotright>= (\<lambda>x. f x \<guillemotright>= g))"
by (cases m) (simp_all add: bind_def)


definition
  catch :: "'e + 'a \<Rightarrow> ('e \<Rightarrow> 'f + 'a) \<Rightarrow> 'f + 'a"
    ("(try(/ _)/ catch(/ _))" [12, 12] 13)
where
 "try m catch f \<equiv> case m of Inl e \<Rightarrow> f e | x \<Rightarrow> x"


abbreviation
  update_error :: "'e + 'a \<Rightarrow> ('e \<Rightarrow> 'f) \<Rightarrow> 'f + 'a" (infixl "<+?" 61)
where
  "m <+? f \<equiv> try m catch (\<lambda>x. error (f x))"

abbreviation
  replace_error :: "'e + 'a \<Rightarrow> 'e \<Rightarrow> 'e + 'a" (infixl "<?" 61)
where
  "m <? e \<equiv> m <+? (\<lambda>x. e)"

lemma catch_return: "(try return x catch f) = Inr x" by (simp add: catch_def)

lemma catch_error: "(try error e catch f) = f e" by (simp add: catch_def)

definition
  isOK :: "'e + 'a \<Rightarrow> bool"
where
  "isOK m \<equiv> (case m of Inl e \<Rightarrow> False | Inr x \<Rightarrow> True)"

lemma isOK_E[elim]: "\<lbrakk>isOK m; \<And>x. m = Inr x \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
unfolding isOK_def by (cases m) simp_all

lemma isOK_I[intro]: assumes "m = Inr x" shows "isOK m"
using assms by (simp add: isOK_def)

lemma isOK_return: "isOK (Inr x) = True" by (simp add: isOK_def)

lemma isOK_error: "isOK (error e) = False" by (simp add: isOK_def)

lemma isOK_bind: "isOK (m \<guillemotright>= f) = (isOK m \<and> isOK (f (run m)))"
by (cases m) (simp_all add: error_bind isOK_error return_bind isOK_return)

lemma isOK_update_error: "isOK (m <+? f) = isOK m"
by (cases m) (simp_all add: catch_error isOK_error catch_return isOK_return)

lemma run_bind:
  assumes "isOK m" shows "run (m \<guillemotright>= f) = run (f (run m))"
using assms by (rule isOK_E) (simp add: return_bind)

lemma run_catch:
  assumes "isOK m" shows "run (try m catch f) = run m"
using assms by (rule isOK_E) (simp add: catch_return)

lemmas run_simps = run_bind run_catch

fun
  foldm :: "('a \<Rightarrow> 'b \<Rightarrow> 'e + 'a) \<Rightarrow> 'a \<Rightarrow> 'b list \<Rightarrow> 'e + 'a"
where 
  "foldm f d [] = return d" |
  "foldm f d (x # xs) = do { y \<leftarrow> f d x; foldm f y xs }"

fun
  forallm_index_aux :: "('a \<Rightarrow> nat \<Rightarrow> 'e + unit) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> (('a \<times> nat) \<times> 'e) + unit"
where
  "forallm_index_aux P i [] = return ()" |
  "forallm_index_aux P i (x # xs) = do {
     P x i <+? Pair (x, i);
     forallm_index_aux P (Suc i) xs
   }"

lemma all_length_Cons:
  "(\<forall>i<length (x # xs). P ((x # xs) ! i) i) = (P x 0 \<and> (\<forall>i<length xs. P (xs ! i) (Suc i)))" (is "?A = (?B \<and> ?C)")
proof (intro iffI conjI)
  assume "?A" thus "?B" by force
next
  assume "?A" thus "?C" by force
next
  assume "?B \<and> ?C"
  show "?A"
  proof (intro allI impI)
    fix i assume "i < length (x # xs)"
    thus "P ((x # xs) ! i) i" using `?B \<and> ?C` by (induct i) simp_all
  qed
qed

lemma isOK_forallm_index_aux:
  "isOK (forallm_index_aux P n xs) = (\<forall>i<length xs. isOK (P (xs ! i) (i + n)))"
proof (induct xs arbitrary: n)
  case Nil thus ?case by (simp add: isOK_return)
next
  case (Cons x xs)
  have "(\<forall>i<length (x # xs). isOK (P ((x # xs) ! i) (i + n))) = (isOK (P x n) \<and> (\<forall>i<length xs. isOK (P (xs ! i) (i + Suc n))))"
  unfolding all_length_Cons[of x xs "\<lambda>x y. isOK (P x (y + n))"] by simp
  thus ?case unfolding Cons[of "Suc n", symmetric] by (simp add: isOK_return isOK_update_error isOK_bind)
qed

definition
  forallm_index :: "('a \<Rightarrow> nat \<Rightarrow> 'e + unit) \<Rightarrow> 'a list \<Rightarrow> (('a \<times> nat) \<times> 'e) + unit"
where
  "forallm_index P xs = forallm_index_aux P 0 xs"

lemma forallm_index[fundef_cong]:
  fixes c::"'a \<Rightarrow> nat \<Rightarrow> 'e + unit"
  assumes "\<And>x i. x \<in> set xs \<Longrightarrow> c x i = d x i"
  shows "forallm_index c xs = forallm_index d xs"
proof -
  {
    fix n
    from assms have "forallm_index_aux c n xs = forallm_index_aux d n xs" by (induct xs arbitrary: n) simp_all
  }
  thus ?thesis unfolding forallm_index_def by simp
qed

hide_const forallm_index_aux

lemma isOK_forallm_index:
  "isOK (forallm_index P xs) = (\<forall>i<length xs. isOK (P (xs ! i) i))"
unfolding forallm_index_def isOK_forallm_index_aux by simp

fun
  forallm :: "('a \<Rightarrow> 'e + unit) \<Rightarrow> 'a list \<Rightarrow> ('a * 'e) + unit"
where
  "forallm f [] = return ()" |
  "forallm f (x # xs) = (f x <+? Pair x \<guillemotright> forallm f xs)"

lemma isOK_forallm:
  "isOK (forallm f xs) = (\<forall>x\<in>set xs. isOK (f x))"
by (induct xs) (simp_all add: isOK_return isOK_update_error isOK_bind)

lemma isOK_if_then: "isOK (if b then m else error e) = (b \<and> isOK m)" by (simp add: isOK_error)
lemma isOK_if_else: "isOK (if b then error e else m) = (\<not> b \<and> isOK m)" by (simp add: isOK_error)

lemmas isOK_simps[simp] =
  isOK_error isOK_bind isOK_return isOK_update_error
  isOK_forallm_index
  isOK_forallm
  isOK_if_then isOK_if_else

fun
  sequence :: "('e + 'a) list \<Rightarrow> 'e + 'a list"
where
  "sequence [] = Inr []" |
  "sequence (m # ms) = do {
    x \<leftarrow> m;
    xs \<leftarrow> sequence ms;
    return (x # xs)
  }"


definition
  mapm :: "('a \<Rightarrow> 'e + 'b) \<Rightarrow> 'a list \<Rightarrow> 'e + 'b list"
where
  "mapm f ms = sequence (map f ms)"

lemmas monad_simps[simp] =
  return_bind bind_return error_bind bind_assoc
  catch_return catch_error
  run_simps


end

